eval 'if test -s `dirname $0`/install/perl/bin/perl5; then `dirname $0`/install/perl/bin/perl5 -U $0 ${1+"$@"}; else echo ERROR: This script must be placed in SASROOT and run from there.; fi; exit 0'
        if $running_under_some_shell;
#----------------------------------------------------------------------------
#  Hotfix Cleanup - The hotfix removal script
#  Copyright (c) 2007, SAS Institute Inc., Cary, N.C., USA
#  All rights reserved.
#----------------------------------------------------------------------------

##--------------------------------------------------------------
## NAME:      hotfixcleanup
##
## PURPOSE: A script to roll back post-SP4 hotfixes before or after the
## installation of new versions of the previously hotfixed products.
##
## NOTES:     This is the main top level file for the
##            program.  The top line invokes perl--it is the only
##            place where perl is specifically called.
## WARNING:   This script was originally written to remove hotfixes
##            for ship event 36.  We cannot be certain that future
##            hotfix 2-byte codes will match product 2-byte codes.
##
## USAGE:     If the user passes no arguments, the script will prompt
##            for a sid file, list the hotfixes to be removed and prompt for ARE YOU SURE
##            For testing purposes, the following switches
##            may be passed:
##                -removeall
##                     This removes the need to ask ARE YOU SURE
##
## DESCRIPTION
##        The user must have permissions to write any file within SASROOT.  The
## user places the script into their SASROOT directory either before or after
## installation of SE36 products.  The user sets executable permissions.  The user
## executes the script.
##
##        Like sassetup, the script is both a shell script and a Perl program.  The
## script launches the copy of Perl that resides in SASROOT/install/perl/bin for
## every 9.1.3 UNIX installation.
##
## The script will consult
## the SASROOT/install/admin/history.hotfix file for a list of hotfixes that have
## been added since Service Pack 4.  This file will provide a list of 2-bytes that
## have hotfixes, which the script will cross-reference with a look-up table it contains,
## and the intersection of these two lists will be the two-bytes with
## hotfixes that need removal.
##        The user will be presented with a list of hotfixes to be removed and asked
## if they actually want these hotfixes removed.  If the user responds "yes"
## the hotfixes will be removed.
##
##        FOR EACH HOTFIX TO BE REMOVED - The script searches for all files in the
## SASROOT/hotfix and SASROOT/nls/???/hotfix directories laid down by that hotfix
## by searching for files that have the hotfix name as a suffix.  The script
## performs the following procedure for each file:
##
##        FOR EACH FILE TO BE REMOVED - The script finds any symbolic links in the
## same directory as the file which point to the file and removes them.
## (Because of the nature of TKPATH look-through when an executable or other file
## is removed from the hotfix area, the non-hotfix file it was intended to replace
## is the one that is used by SAS.  Removing these symlinks will have the effect of
## restoring the pre-hotfix functionality.)  Once the symlink file is
## removed, the file it had pointed to is removed.
##
##        FOR EACH HOTFIX TO BE REMOVED (continued) - After dealing with the files
## (and symlinks) laid down in the hotfix area by the hotfix, the script will check the
## corresponding filelist in SASROOT/install/admin/filelists to make sure that all
## the proper files have been removed.  The script will then remove the filelist
## file itself in order to allow IQ to run successfully.  It will modify the
## SASROOT/install/history.hotfix file to comment out the line concerning the
## addition of the hotfix and a comment line announcing the removal of the hotfix.
##
##        Once each eligible hotfix has been rolled back, the script will return a
## "successful completion" message to the user.  And we shall all rejoice.
##
##--------------------------------------------------------------
BEGIN {

   # If we started up somewhere other than SASROOT, then we need to chdir
   # to the SASROOT for all the @INC stuff to work.  This is related
   # to the eval on the top line, which allows the script to find perl
   # using a relative path even if we didn't start up in SASROOT.
   if ($0 =~ m|^(.*)/[^/]+$|) {   # dirname
      chdir("$1");
   }

   # Remove the absolute paths that were put there when perl was compiled.
   # Use relative paths instead.
   grep { s|^.*(/install/perl/.*)$|.$1| } @INC;
   grep { s|^/opt/perl|./install/perl| } @INC;
   push @INC, "./install/pgm";

   # Turn the debugger on if called with a -d
   if ($ARGV[$#ARGV] eq "-d" and not $^P) {
      pop @ARGV;
      $ENV{PERL5LIB} = join ":", @INC;
      exec ("./install/perl/bin/perl5", "-d", $0, @ARGV);
   }

}

umask(022);
$hostname = `hostname`;
chomp($hostname);

use strict;
use constants;
use File::Copy;
use Cwd;

# This script should always be put in sasroot.
my $sasroot = Cwd::cwd();

# Messages are kept separate for later translation.
my %messages = (
    "prompt_for_sid" => "Please enter the location of your SID (SAS Installation Data) file: ",
    "file_not_found" => "There was no file at the location you specified.",
    "invalid_sid" => "The SID file you indicated was not valid.",
    "could_not_open_sid" => "Unable to open/close the SID file you indicated because: ",
    "no_history_hotfix" => "I could not open the SASROOT/install/admin/history.hotfix file.  You may not have applied any hotfixes since service pack 4.  Please scrutinize your SAS IQ report, and contact Tech Support if there are problems.",
    "no_close_history_hotfix" => "I could not close the history.hotfix file properly.",
    "dir_problem" => "There was a problem reading the directory: ",
    "no_unlink" => "Cannot remove this file: ",
    "no_symlink" => "Cannot create a symbolic link at this location: ",
    "no_history_backup" => "Could not back up your history.hotfix file to: ",
    "no_open_log" => "Could not open the log file for writing: ",
    "no_close_log" => "Could not close the log file: ",
    "usage" => "You have supplied invalid arguments.  Please run the script with no arguments.",
    "wanna" => "Do you wish to remove the following hotfixes?\n(If you respond 'yes' but you wish to continue to run an older version of one of these products, you will have to re-apply the hot fix.)  ",
    "are_you_sure" => "Are you sure you wish to remove these hotfixes (yes/no)? ",
    "please_type" => "Please type yes or no: ",
    "no_changes" => "No changes have been made.",
    "no_write_history" => "I could not write to the SASROOT/install/admin/history.hotfix file.",
    "no_write_log" => "I could not write to teh SASROOT/install/admin/logs/hotfixcleanup.log file.",
    "no_actions" => "There are no hotfixes which need to be removed from your installation.  Script completed successfully.",
    "success" => "Your hotfixes have been removed.  Script completed successfully."
    );

# Process command line arguments.
my $removeall = 0;
if (@ARGV != 0) {
    if (@ARGV == 1 and $ARGV[0] eq "-removeall") {
        $removeall = 1;
    } else {
        print $messages{"usage"};
        exit(1);
    }
}

# Get a 2byte/12byte/name hash for each product in the lookup table.
# (used to get this from the sid, but we want to constrain the list of products this works for.)
my @product_hashes = populate_lookup();
#my @product_hashes = get_twelve_bytes_from_sid($sid);


# Read the history.hotfix file into a string.
unless (open(HISTORY, "$sasroot/install/admin/history.hotfix")) {
    print $messages{"no_hotfix_history"} . " $!\n";
    exit(1);
}
my $history_text;
$history_text .= $_ while <HISTORY>;
unless (close HISTORY) {
    print $messages{"no_close_hotfix_history"} . "$! \n";
    exit(1);
}

# Using history.hotfix, add a hotfix name to the product hashes, and return the subset that have hotfixes
my %hotfix_names = get_hotfix_names_hash($history_text, @product_hashes);

if ((keys %hotfix_names) == 0) {
    print $messages{"no_actions"};
    exit(0);
}

# Do you want to remove the following?
print "\n" . $messages{"wanna"} . "\n\n";


# List hotfixes
foreach my $fix (keys %hotfix_names) {
    print "$fix for \"$hotfix_names{$fix}->{name}\"\n";
}

if (not $removeall) {
# ARE YOU SURE
    print "\n" . $messages{"are_you_sure"};
    my $input = "";
    while ($input !~ /^yes$/i) {
        $input = <STDIN>;
        chomp $input;
        if ($input !~ /^yes|no$/i) {
            print $messages{"please_type"};
            $input = "";
        } elsif ($input =~ /^no$/i) {
            print $messages{"no_changes"};
            exit(0);
        }
    }
}

# Open the log file we'll be writing to.
unless (open(LOG, ">>$sasroot/install/admin/logs/hotfixcleanup.log")) {
    print $messages{"no_write_log"} . " $!\n";
    exit(1);
}

print LOG "hotfixcleanup begins " . timestamp() . "\n";

# Back up the history.hotfix file.
backup_history();

print LOG "history file backed up " . timestamp() . "\n";

# For each hotfix to be removed
foreach my $hotfix_name (keys %hotfix_names) {
    my @removed_files = (); # a list to check with the filelists
    my $hotfix_extension = lc(substr($hotfix_name, 0, 6)); # the 6 character suffix on the files we need to remove for this hotfix
    print LOG "Removing Hotfix: $hotfix_name with suffix $hotfix_extension" . timestamp() . "\n";
    my @paths = ("$sasroot/hotfix");
    push(@paths, <$sasroot/nls/*/hotfix>);

    # Read the directory tree in the hotfix/ and nls/*/hotfix locations for files to be removed.
    # Call remove_file on each file with the hotfix name in its extension
    foreach my $file (@paths) {
        if (-d $file) {
            unless (opendir(DIR, $file)) {
                print $messages{"dir_problem"} . "$file $!";
                print LOG $messages{"dir_problem"} . "$file $! " . timestamp() . "\n";
                exit(1);
            }
            foreach my $subfile (readdir(DIR)) {
                if ($subfile !~ /^\.\.?$/) {
                    if (-d "$file/$subfile") {
                        push(@paths, "$file/$subfile");
                    } elsif ($subfile =~ /$hotfix_extension/) {
                        remove_file("$file/$subfile");
                        push(@removed_files, "$file/$subfile");
                    }
                }
            }
            unless (closedir DIR) {
                print $messages{"dir_problem"} . "$file $!";
                print LOG $messages{"dir_problem"} . "$file $! " . timestamp() . "\n";
                exit(1);
            }
        }
    }

    # Check the filelists
    my @filelists = glob "$sasroot/install/admin/filelists/" . uc($hotfix_extension) . "*";
    my @removable_files = ();
    foreach my $filelist (@filelists) {
        push(@removable_files, removables_from_filelist($filelist));
    }
    foreach my $removable (@removable_files) {
        if (not grep { $_ eq $removable } @removed_files) {
            print LOG "WARNING!! SHOULD HAVE BEEN REMOVED: $removable " . timestamp() . "\n";
        }
    }

    # Remove the filelists
    foreach my $filelist (@filelists) {
        print LOG "REMOVE: $filelist" . timestamp() . "\n";
        if (not unlink($filelist)) {
            print $messages{"no_unlink"} . "$filelist\n$!\n\n";
            print LOG $messages{"no_unlink"} . "$filelist $! " . timestamp() . " \n";
        }
    }

    # Update the text to drop into history.hotfix
    my $two_byte = $hotfix_names{$hotfix_name}->{"two_byte"};
    my $timestamp = timestamp();
    $history_text =~ s/\n(ADD (9.1.3 $two_byte \S+ \S+ $hotfix_name .*Successfully).*?)(\n|$)/\n\# $1\n\# REMOVE $2\# timestamp: $timestamp\n/;
} # end of foreach hotfix to be removed

# Update history.hotfix
unless (open(HISTORY, ">$sasroot/install/admin/history.hotfix")) {
    print $messages{"no_write_history"} . " $!\n";
    print LOG $messages{"no_write_history"} . " $! " . timestamp() . "\n";
    exit(1);
}
print HISTORY $history_text;
unless (close HISTORY) {
    print $messages{"no_write_history"} . " $!\n";
    print LOG $messages{"no_write_history"} . " $! " . timestamp() . "\n";
    exit(1);
}

print LOG "hotfixcleanup completed successfully. " . timestamp() . "\n\n";
print $messages{"no_write_log"} . " $!\n" unless close LOG;
print $messages{"success"};
exit(0);
###########################################################################################################
############################## END OF MAIN EXECUTION ######################################################
###########################################################################################################

##############################################
# NAME: timestamp
# PURPOSE: return a date and time string
#   e.g. Mon 12 7 12:53:23 2007
################################################
sub timestamp {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    my @months = qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/;
    $mon = $months[$mon];
    $year += 1900;
    return "$wday $mon $mday $hour:$min:$sec $year";
}


###############################################
# NAME: backup_history
# PURPOSE: copy the history.hotfix file to
#     sasroot/install/admin/backups with a proper dated
#     file extension.
###############################################
sub backup_history {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    $year -= 100;
    $year = "0" . $year if $year < 10;
    $mon += 1;
    $mon = "0" . $mon if $mon < 10;
    $mday = "0" . $mday if $mday < 10;
    my $prefix = "history.hotfix_$year$mon$mday";
    my @files = <$sasroot/install/admin/backups/$prefix.*>;
    my $number = "000";

    # Look for files from the same date, and count up
    # to an unused extension.
    if (@files) {
        $number = "0";
        foreach my $file (@files) {
            $file =~ /^.*\.(\d*)/;
            $number = $1 if $1 > $number;
        }
        $number = $number + 1;
        if ($number < 10) {
            $number = "00$number";
        } elsif ($number < 100) {
            $number = "0$number";
        }
    }
    my $filename = "$prefix.$number";

    $filename = "$sasroot/install/admin/backups/$filename";
    if (not copy("$sasroot/install/admin/history.hotfix", $filename)) {
        print $messages{"no_history_backup"} . "$filename $!";
        exit(1);
    }

}



###############################################
# NAME: removables_from_filelist
# PURPOSE: Take the path to a filelist as an argument.  Read
#     the filelist.  If a file on the list has a checksum
#     (i.e. it is not a directory) and is either ./hotfix/*
#     or ./nls/*/hotfix/* then return it as part of a list.
###############################################
sub removables_from_filelist {
    my $filelist = shift;
    my $text = "";
    my @retval = ();
    open(FILE, $filelist) or return ();
    while ($text = <FILE>) {
        # some filelists have colons, and some have spaces.  We'll standardize them.
        $text =~ s/:/ /g;
        my @parts = split(" ", $text);
        # The first part is the path to the file with . instead of SASROOT
        my $path = $parts[0];
        # The last part is the checksum for the file, 0 for a directory.
        my $checksum = $parts[-1];
        if ($checksum ne "0" and $path =~ /^\.\/(nls\/[^\/]+\/)?hotfix\//) {
            $path =~ s/^\./$sasroot/;
            push(@retval, $path);
        }
    }
    close FILE;
    return @retval;
}

###############################################
# NAME: remove_file
# PURPOSE: For the file path given via argument, search the directory.
#          If there is a symlink in that directory pointing to this
#          file, we remove the symlink and remove
#          the original file we were passed.
###############################################
sub remove_file {
    my $file = shift;
    my $directory = $file;
    $directory =~ s/(.*)\/.+/$1/;

    # Go through the files in the same directory as the one to be deleted.
    # Find the symlink that points to the file to be deleted.
    my $symlink = "";
    my $prefix = $file; # The file name of the file we're deleting minus it's last dot and  hotfix name
    $prefix =~ s/(.+)\..+/$1/;
    unless (opendir(SUBDIR, $directory)) {
        print $messages{"dir_problem"} . "$directory $!";
        print LOG $messages{"dir_problem"} . "$directory $! " . timestamp() . "\n";
        exit(1);
    }

    foreach my $sibling (grep { $_ !~ /^\.\.?$/ } readdir(SUBDIR)) {
        $sibling = "$directory/$sibling";
        # Find the symlink to redirect
        if (-l $sibling and "$directory/" . readlink($sibling) eq $file) {
            $symlink = $sibling;
        }
    }


    if ($symlink) {
       print LOG "REMOVE SYMLINK: $symlink " . timestamp() . "\n";
       if (not unlink($symlink)) {
           print $messages{"no_unlink"} . "$symlink\n$!\n\n";
           print LOG $messages{"no_unlink"} . "$symlink $! " . timestamp() . "\n";
       }
    }

    print LOG "REMOVE: $file" . timestamp() . "\n";
    if (not unlink($file)) {
        print $messages{"no_unlink"} . "$file\n$!\n\n";
        print LOG $messages{"no_unlink"} . "$file $! " . timestamp() . "\n";
    }
    unless (closedir SUBDIR) {
        print $messages{"dir_problem"} . "$directory $!";
        print LOG $messages{"dir_problem"} . "$directory $! " . timestamp() . "\n";
        exit(1);
    }
}




###############################################
# NAME: get_twelve_bytes_from_sid
# PURPOSE: prompt for sid, read sid, return an array of hashes.
#  Each hash has a "two_byte", "twelve_byte", and "name"
# WARNING: This code is no longer used
###############################################
sub get_twelve_bytes_from_sid {
    my $arg = shift;
    my $sid_path = "";
    my @twelve_bytes = ();
    my @retval = ();

    # Keep looping until we have a proper sid file.
    while (not $sid_path) {

        # If we got a sid on the command line, use it.
        if ($arg and -f $arg) {
            $sid_path = $arg;
            # set this to nothing, so if it was wrong we won't loop forever.
            $arg = "";
        } else {
            # Ask for the sid file
            print $messages{"prompt_for_sid"};
            $sid_path = <STDIN>;
            chomp $sid_path;
            $sid_path =~ s/\s*(.+)\s*/$1/;
        }

        # Check for file availability.
        if (not -f $sid_path) {
            print $messages{"file_not_found"} . "\n";
            $sid_path = "";
        } else {

            # Check to make sure you can open the file.
            if (not open(SID, "$sid_path")) {
                print $messages{"could_not_open_sid"} . $! . "\n";
                $sid_path = "";

                # SID success!
            } else {
                my $mode = "find_file_data";

                # Read the file line by line.
                while (<SID>) {
                    # First, look for the [_FileData_] tag
                    if ($mode eq "find_file_data") {
                        if (/\[_FileData_\]/) {
                            $mode = "find_begin_stream";
                        }
                        # Then look for the ---- BEGIN ---- line
                    } elsif ($mode eq "find_begin_stream") {
                        if (/\$_Stream=-+ BEGIN -+/){
                            $mode = "get_data";
                        }
                        # Then grab all the twelve-bytes
                    } elsif ($mode eq "get_data") {
                        # If you get the ------ END ---- line, you're done.
                        if (/\$_Stream=-+ END -+/) {
                            last;
                            # Ignore the release... line, empty lines, and commented-out lines
                        } elsif (not /^release:/ and not /^#/ and not /^\s+$/) {
                                 my @strings = split(":", $_);

                                 # If there aren't exactly 12 values, we're dealing with an invalid sid
                                 if (@strings != 12) {
                                     print $messages{"invalid_sid"};
                                     exit(1);
                                 }

                                 # Put any PRODUCT (non-bundle) on the list as long as it's not already there.
                                 unless ($strings[1] =~ /^SBUNDLE/ or grep { $_ eq $strings[1] } @twelve_bytes) {
                                     my $twelve_byte = $strings[1];
                                     my $two_byte = substr($strings[10], 0, 2);
                                     my $name = $strings[3];
                                     push(@twelve_bytes, $twelve_byte) ;
                                     push(@retval, {"twelve_byte"=>$twelve_byte, "two_byte"=>$two_byte, "name"=>$name});
                                 }
                            }
                    }
                } # End of while (<SID>)
                close SID or print $messages{"could_not_open_sid"} . $! . "\n" and exit(1);
            } # End of "else I properly opened the SID file"
        } # end of "else $sid_path was fine"

    } # end of "while not $sid_path"
    return @retval;
}






###############################################
# NAME: get_hotfix_names_hash
# PURPOSE: takes the text of the history.hotfix file
#   Returns a hash of hotfix-name keys to product hash values.
#   Each product hash contains name, 12_byte, and 2_byte
###############################################
sub get_hotfix_names_hash {
    my ($history_text, @product_hashes) = @_;
    my %retval = ();

    $history_text =~ /# SERVICE PACK: 4/g;

    # After the SP4 line, we want to pick up successfully installed hotfixes.
    while ($history_text =~ /\nADD 9.1.3 (..) \S+ \S+ (\S+) .*Successfully/g) {
        my $two_byte = $1;
        my $hotfix_name = $2;
        my @found = grep { $_->{"two_byte"} eq $two_byte } @product_hashes;
        if (@found) {
            $retval{$hotfix_name} = $found[0];
        }
    }

    return %retval;
}

###################################
# NAME: populate_lookup
# PURPOSE: return a hash structure for the products eligible for hotfix removal.
###################################
sub populate_lookup {
    return (
        {
            'two_byte' => 'HP',
            'name' => 'SAS High-Performance Forecasting Software',
            'twelve_byte' => 'SASHPF'
        },
        {
            'two_byte' => 'CO',
            'name' => 'Credit Scoring',
            'twelve_byte' => 'CREDITSCRNOD'
        },
        {
            'two_byte' => 'TM',
            'name' => 'SAS Text Miner client Software',
            'twelve_byte' => 'TMINE'
        },
        {
            'two_byte' => 'DA',
            'name' => 'SAS Enterprise Miner',
            'twelve_byte' => 'DMINECLIENT'
        },
        {
            'two_byte' => 'DT',
            'name' => 'SAS Enterprise Miner',
            'twelve_byte' => 'DMINECLIENT'
        }
    );
}


# end of file hotfixcleanup.


